/* GLSL_parallax: Parallax mapping with GLSL shaders
   Copyright (C) 2007 Angelo "Encelo" Theodorou
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glext.h>

#include "init.h"
#include "extra.h"


int main(int argc, char **argv)
{
	SDL_Surface *screen;
	SDL_Event event;
	const SDL_VideoInfo* info = NULL;
	int width = 640;
	int height = 480;
	int bpp = 0;
	int flags = 0;

	const GLubyte* string;
	GLint value;

	struct Material m = {
		{0.6, 0.6, 0.6, 1.0},
		{0.3, 0.3, 0.3, 1.0},
		{100.0}
	};
	struct Light l = {
		{0.2, 0.2, 0.2, 1.0},
		{1.0, 1.0, 1.0, 1.0},
		{1.0, 1.0, 1.0, 1.0},
		{0.0, 0.75, 1.5, 1.0},
	};
	struct Shader sh;
	int withColor = 1;
	int withGloss = 1;
	int withNormal = 1;
	int withParallax = 0;
	GLuint texid[4];
	GLuint vboid[4];

	int quit = 0;
	int withShaders = 0;
	int withBuffers = 0;
	int animateLight = 1;
	int animateCube = 1;
	Uint32 interval;
	GLfloat angleLight = 0.0f;
	GLfloat angleCube = 0.0f;


/* ----- SDL init --------------- */
	if(SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		exit(-1);
	}

	atexit(SDL_Quit);
	
	info = SDL_GetVideoInfo();
	bpp = info->vfmt->BitsPerPixel;

	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

/*	flags = SDL_OPENGL | SDL_FULLSCREEN; */
	flags = SDL_OPENGL | SDL_RESIZABLE;

	
/* ----- Setting up the screen surface --------------- */
	if((screen = SDL_SetVideoMode(width, height, bpp, flags)) == 0) {
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
        	exit(-1);
	}

	SDL_WM_SetCaption("GLSL_parallax by Encelo", NULL);


/* ----- Checking for OpenGL 2 --------------- */
	string = glGetString(GL_VENDOR);
	printf("Vendor: %s\n", string);
	string = glGetString(GL_RENDERER);
	printf("Renderer: %s\n", string);
	string = glGetString(GL_VERSION);
	printf("OpenGL Version: %s\n", string);
	if (string[0] != '2') {
		printf("OpenGL 2 not supported!\n");
		exit(-1);
	}
	string = glGetString(GL_SHADING_LANGUAGE_VERSION);
	printf("GLSL Version: %s\n", string);


/* ----- OpenGL init --------------- */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	/* The following function replaces gluPerspective */
	perspectiveGL(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 100.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_DEPTH_TEST);

	InitLighting(&l, &m);
	InitShaders(&sh);
	InitTextures(texid);
	InitVBOs(vboid);

	glUniform1i(sh.loc0, withColor);
	glUniform1i(sh.loc1, withGloss);
	glUniform1i(sh.loc2, withNormal);
	glUniform1i(sh.loc3, withParallax);

	glActiveTexture(GL_TEXTURE1);
	glDisable(GL_TEXTURE_2D);
	glActiveTexture(GL_TEXTURE2);
	glDisable(GL_TEXTURE_2D);


/* ----- Event cycle --------------- */
	while (!quit) {
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_QUIT:
				quit = 1;
				break;
			case SDL_KEYDOWN:
				switch(event.key.keysym.sym) {
				case SDLK_ESCAPE:
				case SDLK_q:
					quit = 1;
					break;
				case SDLK_l:
					animateLight = !animateLight;
					break;
				case SDLK_k:
					animateCube = !animateCube;
					break;
				case SDLK_v:
					withBuffers = !withBuffers;
					if (withBuffers)
						printf("Using VBOs.\n");
					else
						printf("Not using VBOs.\n");
					break;
				case SDLK_s:
					withShaders = !withShaders;
					if (withShaders)
						printf("Using shaders.\n");
					else
						printf("Not using shaders.\n");
					break;
				case SDLK_c:
					withColor = !withColor;
					if (withColor) {
						printf("Using Color Mapping.\n");
						glActiveTexture(GL_TEXTURE0);
						glEnable(GL_TEXTURE_2D);
					}
					else {
						printf("Not using Color Mapping.\n");
						glActiveTexture(GL_TEXTURE0);
						glDisable(GL_TEXTURE_2D);
					}
					glUniform1i(sh.loc0, withColor);
					break;
				case SDLK_g:
					withGloss = !withGloss;
					if (withGloss)
						printf("Using Specular Mapping.\n");
					else
						printf("Not using Specular Mapping.\n");
					glUniform1i(sh.loc1, withGloss);
					break;
				case SDLK_n:
					withNormal = !withNormal;
					if (withNormal)
						printf("Using Normal Mapping.\n");
					else
						printf("Not using Normal Mapping.\n");
					glUniform1i(sh.loc2, withNormal);
					break;
				case SDLK_p:
					withParallax = !withParallax;
					if (withParallax)
						printf("Using Parallax Mapping.\n");
					else
						printf("Not using Parallax Mapping.\n");
					glUniform1i(sh.loc3, withParallax);
					break;
				case SDLK_F1:
					SDL_WM_ToggleFullScreen(screen);
					break;
				default:
					break;
				}
				break;
			case SDL_VIDEORESIZE:
				if((screen = SDL_SetVideoMode(event.resize.w, event.resize.h, bpp, flags)) == 0) {
					fprintf(stderr, "Video resize failed: %s\n", SDL_GetError());
					exit(-1);
				}
				glPushAttrib(GL_TRANSFORM_BIT);
				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				perspectiveGL(45.0f, (GLfloat)event.resize.w/(GLfloat)event.resize.h, 0.001f, 100.0f);
				glViewport(0.0f, 0.0f, event.resize.w, event.resize.h);	
				glPopAttrib();
				break;
			case SDL_ACTIVEEVENT:
				if(event.active.state != SDL_APPMOUSEFOCUS && event.active.gain == 0)
					while(1){
						SDL_WaitEvent(&event);
						if(event.type == SDL_ACTIVEEVENT && 
							event.active.state != SDL_APPMOUSEFOCUS && 
							event.active.gain == 1);
							break;
					}
				break;
			default:
				break;
			}
		}

		interval = FrameTiming();

		
/* ----- Blitting on the screen --------------- */
		if (withShaders)
			glUseProgram(sh.p);
		else
			glUseProgram(NULL);

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		if (animateCube)
			angleCube += 0.1f * interval;
		if (animateLight)
			angleLight += 0.001 * interval;

		l.position[0] = sinf(angleLight) * 1.0f;
		l.position[1] = sinf(angleLight) * 1.0f;
		glLoadIdentity();
		glTranslatef(0.0f, 0.0f, -6.0f);
		glRotatef(15.0f, 1.0f, 0.0f, 0.0f);

		if (withShaders)
			glUseProgram(NULL);
		UpdateLight(&l);
		if (withShaders)
			glUseProgram(sh.p);

		glRotatef(angleCube, 0.0f, 1.0f, 0.0f);

		DrawCube(withBuffers, vboid);

		SDL_GL_SwapBuffers();		
		/* SDL_Delay(20); */ /* Decomment this if you want 1/50th screen update */
	}

	
/* ----- Quitting --------------- */
	glDeleteBuffers(4, vboid);
	glDeleteTextures(4, texid);

	glDetachShader(sh.p, sh.v);
	glDetachShader(sh.p, sh.f);
	glDeleteShader(sh.v);
	glDeleteShader(sh.f);
	glDeleteProgram(sh.p);

	SDL_Quit();
	return 0;	
}
